/*
 * Created on Apr 7, 2005
 *
 */
package com.sunlabs.archsim.fleet.HornAndFunnel;

import java.util.Map;
import java.util.HashMap;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.gasp.Branch;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author Igor
 *
 * Model of a Fleet instruction Horn.
 * Interface: in, ack in on the input side,
 * out[i], ackOut[i] on the outputr side, where
 * i ranges from 0 to (2^depth)-1.
 */
public class InstructionHorn extends Component {

	
	/** the number of internal branching levels */
	int depth;
	
	/** the number of out ports - on the output side */
	int numOutputs;
	
	final int branchFactor= 2;
	
	protected Map<String, Terminal> nameToTerminalMap;
	
	Branch[][] branch;
	InstructionHornBranchAction[][] branchAction;

//	InputTerminal[][] in;
//	OutputTerminal[][] ackIn;

//	OutputTerminal[][] intOut;
//	InputTerminal[][] intAckOut;
		
	State[][] intDataState;
	State[][] intAckState;
	
	// for XML interface, needs improvement
	public InstructionHorn(String name, Parameters params) {
		super(name, null);
		
		int depthParam= hornAndFunnelDefaults.DEPTH_DEF;
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(hornAndFunnelDefaults.DEPTH) 
						|| pName.equals(FleetGlobal.PARAMETER) ) {
					try {
						depthParam= Integer.parseInt(sValue);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ hornAndFunnelDefaults.DEPTH
								+ ": " + sValue
								+ ". Value is IGNORED, default value of "
								+ hornAndFunnelDefaults.DEPTH_DEF 
								+ " is used.");						
					}
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				}
				
			}
		} // if params != null		
		
		buildInstructionHorn(depthParam);		
	}
	
	public InstructionHorn(String name, int depth, Parameters params) {
		super(name, null);
		
		int depthParam= depth;
		
		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(hornAndFunnelDefaults.DEPTH)) {
					// ignore depth
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				}
				
			}
		} // if params != null		
		
		buildInstructionHorn(depthParam);		
	}
	
	
	/**
	 * Make an instruction Horn. Has 2^depth outputs.
	 * @param name name
	 * @param d = depth the number of levels inside the horn
	 */
	public InstructionHorn(String name, int d) {
		super(name, null);
		
		buildInstructionHorn(d);

	} // InctructionHornConstructor
	
	
	protected void buildInstructionHorn(int d) {
		depth= d;
		numOutputs= 1 << depth;
		
		//prepare arrays for required components
		// branches
		branch= new Branch[depth][];
		// branchActions for the branches
		branchAction= new InstructionHornBranchAction[depth][];
		
		intDataState= new State[depth][];
		intAckState= new State[depth][];
		
		// mask used to steer data
		// set the bit that decides the direction to 1, the rest to 0
		// start by setting the high-order bit to 1
		int mask= (1 << (depth-1));

		// number of components per level
		// initialize to 1 for the firts level (index 0)
		// this number is doubled in each iteration
		int perLevel= 1;
		
		// create components for each level
		for (int level=0; level<depth; level++) {
			// prepare arrays for components in this level
			branch[level]= new Branch[perLevel];
			branchAction[level]= new InstructionHornBranchAction[perLevel];

			// these states lead to the previous level
			intDataState[level]= new State[perLevel];
			intAckState[level]= new State[perLevel]; 
			
			
			
			// make components and connections for this level
			for (int i=0; i< perLevel; i++) {
				branchAction[level][i]= new InstructionHornBranchAction(
						FleetGlobal.DELIMITER+level
						+FleetGlobal.DELIMITER+i, mask);
				branch[level][i]= new Branch(
						FleetGlobal.BRANCH+FleetGlobal.DELIMITER
						+level+FleetGlobal.DELIMITER+i, 
						branchFactor, branchAction[level][i]);
				this.addMember(branch[level][i]);
				
				// if this is not the first branch, establish the connection 
				if (level != 0) {
					// make the states
					intDataState[level][i]= new State(
							FleetGlobal.DATA_STATE+FleetGlobal.DELIMITER
							+level+FleetGlobal.DELIMITER+i, this);
					intAckState[level][i]= new State(
							FleetGlobal.ACK_STATE+FleetGlobal.DELIMITER
							+level+FleetGlobal.DELIMITER+i, this);
					intAckState[level][i].setInitiallyActive(null);
					
					// attach the terminals
					intDataState[level][i].attach(
							branch[level-1][i/2].getOut(i % branchFactor));
					intDataState[level][i].attach(
							branch[level][i].getIn());
					
					intAckState[level][i].attach(
							branch[level-1][i/2].getAckOut(i % branchFactor));
					intAckState[level][i].attach(
							branch[level][i].getAckIn());
					
					// all of the internal ack states are active!
					// this state was constructed as initially active
					// intAckState[level][i].setActive(null);
					
				} // if level != 0 -> make internal states 
				  // and corresponding connections
				
			} // for i = 0..perLevel
			// next level branches according to the branchFactor
			perLevel= perLevel*branchFactor;
			// prepare the mask for the next level
			mask= mask >> 1;
		} // for level
		
		// set terminal names and establish name-to-terminal mapping
		nameToTerminalMap=  new HashMap<String, Terminal>();
		Terminal t;
		
		// get input - at ther root of the horn
		t= getIn();
		// set the name
		t.setAlias(FleetGlobal.IN);
		// establish the mapping
		nameToTerminalMap.put(FleetGlobal.IN, t);
		nameToTerminalMap.put(t.getBaseName(), t);

		// the same for the ackIn from root
		t= getAckIn();
		t.setAlias(FleetGlobal.ACK_IN);
		nameToTerminalMap.put(FleetGlobal.ACK_IN, t);
		nameToTerminalMap.put(t.getBaseName(), t);
		
		for (int i=0; i< numOutputs; i++) {
			// set the name and establish the mapping for the output side
			t= getOut(i);
			t.setAlias(FleetGlobal.OUT+i);
			nameToTerminalMap.put(FleetGlobal.OUT+i, t);
			nameToTerminalMap.put(t.getBaseName(), t);
			
			t= getAckOut(i);
			t.setAlias(FleetGlobal.ACK_OUT+i);
			nameToTerminalMap.put(FleetGlobal.ACK_OUT+i, t);
			nameToTerminalMap.put(t.getBaseName(), t);
		} // for
				
	} // buildInstructionHorn
	
	public InputTerminal getIn() {
		return branch[0][0].getIn();
	}
	
	public OutputTerminal getAckIn() {
		return branch[0][0].getAckIn();
	}
	
	public OutputTerminal getOut(int n) {
		if ( 0 <= n && n < numOutputs)
			return branch[depth-1][n/branchFactor].getOut(n%branchFactor);
		else
			throw new IndexOutOfBoundsException();
	}
	
	public InputTerminal getAckOut(int n) {
		if ( 0 <= n && n < numOutputs)
			return branch[depth-1][n/branchFactor].getAckOut(n%branchFactor);
		else
			throw new IndexOutOfBoundsException();	
	}
	
	/** return a terminal from the external boundary
	 *  given the terminal or alias.
	 */
	public Terminal getTerminal(String name) {
		Object o= nameToTerminalMap.get(name);
		if (o instanceof Terminal)
			return (Terminal)o;
		else 
			return null;
	} // get erminal
	
} // InstructionHorn
